@@ -1,27 +1,15 @@
 import json
-import os
 import threading
-from functools import lru_cache
-from io import BytesIO
+import pylru
 
 import logging
-import pycurl
+import urllib3
 
-__version__ = "2.1.0"
+__version__ = "2.2.0"
 __client_version__ = "wurfl-microservice-python_%s" % __version__
-__default_http_timeout__ = 10000
+__default_http_timeout__ = 10
 config_path = ""
 
-
-def read_cache_size():
-    default_cache_size = 200000
-    if "WM_CACHE_SIZE" in os.environ:
-        return int(os.environ["WM_CACHE_SIZE"])
-    return default_cache_size
-
-
-cache_size = read_cache_size()
-
 logger = logging.getLogger("wurfl-microservice")
 
 
@@ -49,6 +37,9 @@ def to_lower_keys_dict(headers):
 class WmClient:
 
     def __init__(self):
+        self.staticCaps = []
+        self.virtualCaps = []
+        self.ltime = None
         self.scheme = "http"
         self.host = ""
         self.port = 80
@@ -60,13 +51,16 @@ class WmClient:
         self.device_os_lock = threading.Lock()
         self.device_makes_lock = threading.Lock()
         self.device_OSes = []
-        self.curl_post = pycurl.Curl()
-        self.curl_post.setopt(pycurl.TIMEOUT_MS, __default_http_timeout__)
-        self.curl_post.setopt(pycurl.HTTPHEADER, ['Accept: application/json',
-                                                  'Content-Type: application/json'])
-        self.curl_get = pycurl.Curl()
-        self.curl_get.setopt(pycurl.TIMEOUT_MS, __default_http_timeout__)
-        self.curl_get.setopt(pycurl.HTTPHEADER, ['Accept: application/json'])
+        self.head_cache_lock = threading.Lock()
+        self.dev_cache_lock = threading.Lock()
+        self.make_models = []
+        self.device_makes = []
+        self.device_os_versions_dict = dict()
+        self.internal_client = urllib3.PoolManager(num_pools=200, maxsize=200, timeout=__default_http_timeout__)
+
+        # caches: by default client has none. Yet it is strongly recommended to set them up.
+        self.dev_id_cache = None
+        self.head_cache = None
 
     @staticmethod
     def create(scheme, host, port, baseURI):
@@ -83,60 +77,46 @@ class WmClient:
         client.important_headers = info.important_headers
         client.staticCaps = sorted(info.static_capabilities)
         client.virtualCaps = sorted(info.virtual_capabilities)
+        client.ltime = info.ltime
 
         return client
 
     def set_http_timeout(self, timeout):
-        """Sets HTTP connection timeout in milliseconds"""
+        """Sets HTTP connection timeout in seconds"""
         if timeout is None or timeout <= 0:
-            self.curl_post.setopt(pycurl.TIMEOUT_MS, __default_http_timeout__)
-        else:
-            self.curl_post.setopt(pycurl.TIMEOUT_MS, timeout)
-
-        if timeout is None or timeout <= 0:
-            self.curl_get.setopt(pycurl.TIMEOUT_MS, __default_http_timeout__)
-        else:
-            self.curl_get.setopt(pycurl.TIMEOUT_MS, timeout)
+            self.internal_client = urllib3.PoolManager(num_pools=200, maxsize=200, timeout=timeout)
 
     def destroy(self):
         """Closes and deallocates all resources used to connect to server.
         Function calls made after this one will cause error"""
-        self.clear_cache()
-
-        self.curl_get.close()
-        self.curl_get = None
-
-        self.curl_post.close()
-        self.curl_post = None
+        self.clear_caches()
+        self.internal_client = None
 
     def get_info(self):
         """Returns info about WURFL microservice server.
         It raises a WmClientError in case of connection errors/timeouts"""
 
         msg = "Client creation failed. Unable to connect to WURFL microservice server - "
-        data = BytesIO()
+
         try:
-            # performs a GET using pycurl
-            self.curl_get.setopt(self.curl_get.URL, self.__create_URL("/v2/getinfo/json"))
-            self.curl_get.setopt(self.curl_get.WRITEFUNCTION, data.write)
-            self.curl_get.perform()
-            res_code = self.curl_get.getinfo(pycurl.HTTP_CODE)
+            # performs a GET and get status again
+            res = self.internal_client.request("GET", self.__create_URL("/v2/getinfo/json"))
+            res_code = res.status
 
-        except ConnectionError as e:
-            if hasattr(e, 'message'):
-                msg += e.message
+            if 200 <= res_code < 400:
+                info = json.loads(res.data.decode('utf-8'))
+                res.release_conn()
+                self.clear_cache_if_needed(info["ltime"])
+                return JsonInfoData(info)
+            else:
+                msg = "get_info - Unable to get WURFL microservice server info - response code: " + str(res_code)
+                logging.error(msg)
                 raise WmClientError(msg)
-        except Exception:
-            raise WmClientError(msg)
 
-        if 200 <= res_code < 400:
-            info = json.loads(data.getvalue().decode(encoding='UTF-8'))
-            data.close()
-            return JsonInfoData(info)
-        else:
-            msg = "get_info - Unable to get WURFL microservice server info - response code: " + str(res_code)
-            logging.error(msg)
-            raise WmClientError(msg)
+        except Exception as e:
+            err_msg_tpl = "An exception of type {0} occurred. Arguments:\n{1!r}"
+            message = err_msg_tpl.format(type(e).__name__, e.args)
+            raise WmClientError(message)
 
     def __create_URL(self, path):
         url = self.scheme + "://" + self.host
@@ -159,7 +139,7 @@ class WmClient:
         """sets the list of static capabilities handled by this client"""
         if capsList is None:
             self.requested_static_caps = None
-            self.clear_cache()
+            self.clear_caches()
             return
 
         stCaps = []
@@ -168,13 +148,13 @@ class WmClient:
                 stCaps.append(name)
 
         self.requested_static_caps = stCaps
-        self.clear_cache()
+        self.clear_caches()
 
     def set_requested_virtual_capabilities(self, capsList):
         """sets the list of virtual capabilities handled by this client"""
         if capsList is None:
             self.requested_virtual_caps = None
-            self.clear_cache()
+            self.clear_caches()
             return
 
         vCaps = []
@@ -183,14 +163,14 @@ class WmClient:
                 vCaps.append(name)
 
         self.requested_virtual_caps = vCaps
-        self.clear_cache()
+        self.clear_caches()
 
     def set_requested_capabilities(self, capsList):
         """sets the list of virtual and static capabilities handled by this client"""
         if capsList is None:
             self.requested_static_caps = None
             self.requested_virtual_caps = None
-            self.__internal_request.cache_clear()
+            self.clear_caches()
             return
 
         capNames = []
@@ -208,45 +188,137 @@ class WmClient:
 
         if len(vcapNames) > 0:
             self.requested_virtual_caps = vcapNames
-        self.__internal_request.cache_clear()
+        self.clear_caches()
+
+    def set_cache_size(self, size):
+        """
+        Creates a new LRU cache with the given size or, if cache has already been created, increases the size to
+        given value.
+        :param size: new size of the cache
+        """
+        if self.dev_id_cache is None:
+            self.dev_id_cache = pylru.lrucache(10000)
+
+        if self.head_cache is None:
+            self.head_cache = pylru.lrucache(size)
+        else:
+            self.head_cache.size(size)
+
+    def clear_caches(self):
+        if self.head_cache is not None:
+            with self.head_cache_lock:
+                self.head_cache.clear()
+
+        if self.dev_id_cache is not None:
+            with self.dev_cache_lock:
+                self.dev_id_cache.clear()
+
+        self.make_models = []
+        self.device_makes = []
+
+        with self.device_os_lock:
+            self.device_OSes = []
+            self.device_os_versions_dict = dict()
+
+    def clear_cache_if_needed(self, ltime):
+        if ltime is not None and not (ltime == self.ltime):
+            self.ltime = ltime
+            self.clear_caches()
+
+    def get_actual_cache_size(self):
+
+        csize = [0] * 2
+
+        with self.dev_cache_lock:
+            if self.dev_id_cache is not None:
+                csize[0] = len(self.dev_id_cache)
+
+        with self.head_cache_lock:
+            if self.head_cache is not None:
+                csize[1] = len(self.head_cache)
+
+        return csize
+
+    def safe_put_device(self, cache_type, key, device):
+        """
+        Puts the device in the cache corresponding to the key value if it has been initialized
+        :param cache_type: the type of cache to which save the element
+        :param key: the cache type
+        :param device: the device to store in cache
+        """
+
+        if cache_type == HEADERS_CACHE_TYPE:
+            if self.head_cache is not None:
+                with self.head_cache_lock:
+                    self.head_cache[key] = device
+
+        elif cache_type == DEVICE_ID_CACHE_TYPE:
+            if self.dev_id_cache is not None:
+                with self.dev_cache_lock:
+                    self.dev_id_cache[key] = device
 
-    @lru_cache(maxsize=cache_size)
     def __internal_request(self, path, req):
 
-        device = None
-        # Buffer where data sent from the server are written
-        data = BytesIO()
-        url = self.__create_URL(path)
-        try:
+        if req is None:
+            raise WmClientError("Request object cannot be None")
 
-            self.curl_post.setopt(self.curl_post.URL, url)
+        cache_key = None
+        device = None
+        if DEVICE_ID_CACHE_TYPE == req.cache_type:
+            cache_key = req.wurfl_id
+        elif HEADERS_CACHE_TYPE == req.cache_type:
+            cache_key = req.get_headers_cache_key()
+
+        # First, do a cache lookup
+        if req.cache_type is not None and len(req.cache_type) > 0 and cache_key is not None and len(cache_key) > 0:
+            if req.cache_type == DEVICE_ID_CACHE_TYPE and self.dev_id_cache is not None:
+                with self.dev_cache_lock:
+                    device = self.dev_id_cache.get(req.wurfl_id)
+                if device is not None:
+                    return device
+            elif req.cache_type == HEADERS_CACHE_TYPE and self.head_cache is not None:
+                with self.head_cache_lock:
+                    device = self.head_cache.get(cache_key)
+
+        if device is not None:
+            return device
+
+        # No device found in cache, do a server lookup
+
+        try:
             # prepares request payload
-            d = json.dumps(req.__dict__)
-            payload = d.encode()
-            # Sets request method to POST,
-            # and data to send in request body.
-            self.curl_post.setopt(self.curl_post.POSTFIELDS, payload)
-            self.curl_post.setopt(self.curl_post.WRITEFUNCTION, data.write)
-            self.curl_post.perform()
-            res_code = self.curl_post.getinfo(pycurl.HTTP_CODE)
+            payload = json.dumps(req.__dict__).encode('utf-8')
+
+            res = self.internal_client.request(
+                'POST',
+                self.__create_URL(path),
+                body=payload,
+                headers={'Content-Type': 'application/json'})
+
+            res_code = res.status
+
             if 200 <= res_code < 400:
-                json_response = json.loads(data.getvalue().decode(encoding='UTF-8'))
+                json_response = json.loads(res.data.decode(encoding='UTF-8'))
                 device = JsonDeviceData(json_response)
-                data.close()
+                res.release_conn()
+
                 if device.error != "":
                     raise WmClientError("Unable to complete request to WM server: " + device.error)
 
+                # check if cache must be cleared
+                self.clear_cache_if_needed(device.ltime)
+                # add device to cache
+                if device is not None and len(cache_key) > 0:
+                    self.safe_put_device(req.cache_type, cache_key, device)
+
         except Exception as e:
-            logging.error(str(e))
             if isinstance(e, WmClientError):
                 raise e
             else:
                 msg = self.__format_except_message__(e, "Unable to complete request to WM server: ")
                 logging.error(msg)
                 raise WmClientError(msg)
-        finally:
-            # closing buffer
-            data.close()
+
         return device
 
     def lookup_useragent(self, useragent):
@@ -324,18 +396,16 @@ class WmClient:
             if self.device_os_lock.locked():
                 self.device_os_lock.release()
             return
-        data = BytesIO()
+
         try:
             url = self.__create_URL("/v2/alldeviceosversions/json")
-            self.curl_get.setopt(self.curl_get.URL, url)
-            self.curl_get.setopt(self.curl_get.WRITEFUNCTION, data.write)
-            self.curl_get.perform()
-            res_code = self.curl_get.getinfo(pycurl.HTTP_CODE)
+            res = self.internal_client.request("GET", url)
+            res_code = res.status
             if not (200 <= res_code < 400):
                 raise WmClientError("Unable to get device OSes data - response code: " + str(res_code))
             else:
-                strOSes = json.loads(data.getvalue().decode(encoding='UTF-8'))
-                data.close()
+                strOSes = json.loads(res.data.decode(encoding='UTF-8'))
+                res.release_conn()
             devOses = []
             ddict = dict()
             for osVer in strOSes:
@@ -351,13 +421,12 @@ class WmClient:
             if not self.device_os_lock.locked():
                 self.device_os_lock.acquire()
             self.device_OSes = devOses
-            self.deviceOsVersionsMap = ddict
+            self.device_os_versions_dict = ddict
             self.device_os_lock.release()
         except Exception as e:
             msg = self.__format_except_message__(e, "An error occurred getting device os name and version data - {}"
                                                  .format(str(e)))
             logging.error(msg)
-            data.close()
             raise WmClientError(msg)
         finally:
             if self.device_os_lock.locked():
@@ -366,8 +435,8 @@ class WmClient:
     def get_all_versions_for_OS(self, osName):
         """:return a list of all the known versions for the given device OS"""
         self.__load_device_OSes_data()
-        if osName in self.deviceOsVersionsMap:
-            osVers = self.deviceOsVersionsMap[osName]
+        if osName in self.device_os_versions_dict:
+            osVers = self.device_os_versions_dict[osName]
             for ver in osVers:
                 if "" == ver:
                     osVers.remove(ver)
@@ -377,27 +446,24 @@ class WmClient:
             raise WmClientError(msg)
 
     def __load_device_makes_data(self):
-
-        # If deviceMakes cache has values everything has already been loaded, thus we exit
-        self.device_makes_lock.acquire()
-        if len(self.device_makes) > 0:
-            if self.device_makes_lock.locked():
-                self.device_makes_lock.release()
+        try:
+            # If deviceMakes cache has values everything has already been loaded, thus we exit
+            self.device_makes_lock.acquire()
+            if len(self.device_makes) > 0:
                 return
+        finally:
+            self.device_makes_lock.release()
 
-        # No values already loaded, let's do it.
-        data = BytesIO()
+            # No values already loaded, let's do it.
         try:
             url = self.__create_URL("/v2/alldevices/json")
-            self.curl_get.setopt(self.curl_get.URL, url)
-            self.curl_get.setopt(self.curl_get.WRITEFUNCTION, data.write)
-            self.curl_get.perform()
-            res_code = self.curl_get.getinfo(pycurl.HTTP_CODE)
+            res = self.internal_client.request("GET", url)
+            res_code = res.status
             if not (200 <= res_code < 400):
                 raise WmClientError("Unable to get device makers data - response code: " + str(res_code))
             else:
-                localMakeModels = json.loads(data.getvalue().decode(encoding='UTF-8'))
-                data.close()
+                localMakeModels = json.loads(res.data.decode(encoding='UTF-8'))
+                res.release_conn()
 
                 dmMap = dict()
                 devMakes = []
@@ -409,15 +475,11 @@ class WmClient:
                     mdMkNames = dmMap.get(mkModel.brand_name)
                     if mdMkNames is None:
                         mdMkNames = []
-                        dmMap[mkModel.brand_name] = mdMkNames
+                    dmMap[mkModel.brand_name] = mdMkNames
 
                     mdMkNames.append(mkModel)
-                    if not self.device_makes_lock.locked():
-                        self.device_makes_lock.acquire()
                     self.deviceMakesMap = dmMap
                     self.device_makes = devMakes
-                    if self.device_makes_lock.locked():
-                        self.device_makes_lock.release()
         except Exception as e:
             msg = self.__format_except_message__(e, "An error occurred getting makes and model data ")
             logging.error(msg)
@@ -440,12 +502,10 @@ class WmClient:
             raise WmClientError(msg)
 
     def cache_info(self):
-        """:return the current state of WMClient internal cache (hits, misses, max size)"""
-        return self.__internal_request.cache_info()
-
-    def clear_cache(self):
-        """clears WM client internal cache"""
-        return self.__internal_request.cache_clear()
+        """
+        This method is deprecated since version 2.2.0 and will always return None. Use get_actual_cache_size instead
+        :return the current state of WMClient internal cache (hits, misses, max size)"""
+        return None
 
     def get_api_version(self):
         """:return the version of this WM Client"""
@@ -477,6 +537,7 @@ class JsonInfoData:
         self.important_headers = info_dict["important_headers"]
         self.static_capabilities = info_dict["static_caps"]
         self.virtual_capabilities = info_dict["virtual_caps"]
+        self.ltime = info_dict["ltime"]
 
 
 class JsonDeviceData:
@@ -499,18 +560,7 @@ class Request:
         self.important_headers = important_headers
         self.key = None
 
-    def __eq__(self, other):
-        if self is None or other is None:
-            return False
-        if self.cache_type == DEVICE_ID_CACHE_TYPE:
-            return self.wurfl_id == other.wurfl_id
-        else:
-            return self.get_user_agent_cache_key() == other.get_user_agent_cache_key()
-
-    def __hash__(self):
-        return hash(self.get_user_agent_cache_key())
-
-    def get_user_agent_cache_key(self):
+    def get_headers_cache_key(self):
 
         if self.key is not None:
             return self.key
